// Program written by Maarten Meuris AKA Nyerguds
program rtmtype;

uses
  crt,
  dos,
  Math,
  SysUtils,
  StrUtils;

type
  TerrainType =
  Record
    ByteId: Byte;
    Letter:Char;
    Desc: String[65];
  end;

const
  TerrainTypes:array[1..8] of TerrainType =
  (                               //'_________________________________________________________________' 65
    (ByteId: $0; Letter: 'X'; Desc: 'Filler tile, or [Clear] terrain on 1x1 sets with multiple tiles.'),
    (ByteId: $3; Letter: 'C'; Desc: '[Clear] Normal clear terrain.'),
    (ByteId: $6; Letter: 'B'; Desc: '[Beach] Sandy beach. Can''t be built on.'),
    (ByteId: $8; Letter: 'I'; Desc: '[Rock]  Impassable terrain.'),
    (ByteId: $9; Letter: 'R'; Desc: '[Road]  Units move faster on this terrain.'),
    (ByteId: $A; Letter: 'W'; Desc: '[Water] Ships can travel over this.'),
    (ByteId: $B; Letter: 'V'; Desc: '[River] Ships normally can''t travel over this.'),
    (ByteId: $E; Letter: 'H'; Desc: '[Rough] Rough terrain. Can''t be built on.')
  );

var
  // File to edit
  fileobj: file;
  // 0=default 1=fromfile, 2=fromcodes, 3=forceOne, 4=fileinfo, 5=aborted
  RunMode: Integer;
  TerrainOutputFile: String;
  ForceLength: Boolean;
  DefaultFill: byte;
  FixToSingleCell: Boolean;
  // parameter info. Stops being relevant after the internal options are all set.
  TerrainInputFile: String;
  TerrainInputCells: array [0..$FF] of byte;
  TerrainInputCellsLength: Integer;
  TerrainInputString: String;
  TerrainHexReadMode: Boolean;
  DefaultFillInputChar: char;
  DefaultHexReadMode: Boolean;
  // RTM header info
  {Full header:
  0x00 int16 tile_X (always 24)
  0x02 int16 tile_Y (always 24)
  0x04 int32 frames
  0x08 int16 frames_x
  0x0A int16 frames_y
  0x0C int32 file_length
  0x10 int32 offset_endofheader (always 0x28)
  0x14 int32 unkn_00
  0x18 int32 unkn_01
  0x1C int32 offset_footer02
  0x20 int32 offset_footer03_typeframes
  0x24 int32 offset_footer01  // start of footer }

  Cells: Cardinal;
  CellsX, CellsY: Word;
  TypesOffset: Cardinal;


// =========================================
//                 MISCELLANEOUS
// =========================================

// To convert value to string on the fly without needing a variable to store it in.
function toString(b: boolean): string;
begin
  str(byte(b), Result);
end;

function toString(b: byte): string;
begin
  str(b, Result);
end;

function toString(i: integer): string;
begin
  str(i, Result);
end;

function toString(i: cardinal): string;
begin
  str(i, Result);
end;

// Press Key To Continue procedure with buffer clear.
procedure cntin(Msg: string);
var inkey : char;
begin
  if (length(Msg) > 0) then
  begin
    writeln();
    Write(Msg);
  end;
  repeat
  until keypressed;
  inkey := ReadKey();
  if (keypressed and (inkey = #0)) then
    ReadKey(); {Clears the keyboard buffer if a special key was pressed}
  if (length(Msg) > 0) then
    writeln();
end; {cntin}

procedure cntin();
begin
  cntin('');
end; {cntin}

procedure cntin(showMsg: boolean);
begin
  if showMsg then
    cntin('Press any key to continue. . . ')
  else
    cntin('');
end; {cntin}

function hexchar2byte(hex: char; var code: Integer): byte;
begin
  Result := 0;
  if (byte(hex) >= 65) and (byte(hex) <= 70) then
    Result := byte(hex) - 55
  else if (byte(hex) >= 48) and (byte(hex) <= 57) then
    Result := byte(hex) - 48
  else
    code:=1;
end;

function byte2hexchar(hex: byte): char;
begin
  if ((hex mod 16) >= 10) then
    Result := char((hex mod 16) - 10 + $41)
  else
    Result := char((hex mod 16) + $30);

end;

// =========================================
//                 INPUT / OUTPUT
// =========================================

function openFile(filename: string): boolean;
var
  status: integer;
begin
  Assign(fileobj, filename);
{$I-}
  reset(fileobj, 1);
{$I+}
  status := IOResult;
  if status <> 0 then
  begin
    writeln('Error opening "', filename, '".');
    Result := False;
  end
  else
  begin
    Result := True;
  end;
end; {openFile}

procedure closeFile();
begin
  Close(fileobj);
{$I+}
end; {closeFile}

function getFileSize(): Int64;
begin
  Result:=FileSize(fileobj);
end; {getFileSize}


function readByte(address: cardinal): byte;
begin
  if FileSize(fileobj) >= address then
  begin
    Result := 0;
    seek(fileobj, address);
    blockread(fileobj, Result, 1);
  end;
end;

// converts big-endian to little-endian
function readBytes(address: cardinal; nrofbytes: byte): cardinal;
var
  i, cur: byte;
begin
  nrofbytes := min(4, nrofbytes);
  nrofbytes := max(1, nrofbytes);
  nrofbytes := nrofbytes - 1;
  Result := 0;
  for i := 0 to nrofbytes do
  begin
    cur := readByte(address);
    Result := Result + (cur shl (8 * i));
    address := address + 1;
  end;
end;

procedure writeByte(address: cardinal; Data: byte);
begin
  seek(fileobj, address);
  blockwrite(fileobj, Data, 1);
end;

// bytenr: 0 = lowest byte, 3 = highest byte
function getByte(Value: cardinal; bytenr: cardinal): byte;
begin
  bytenr := min(3, bytenr);
  bytenr := max(0, bytenr);
  Value := Value shr (8 * (bytenr));
  Value := Value and $FF;
  Result := byte(Value);
end;

// converts little-endian to big-endian
procedure writeBytes(address, Value: cardinal; nrofbytes: byte);
var
  i: byte;
begin
  nrofbytes := min(4, nrofbytes);
  nrofbytes := max(1, nrofbytes);
  for i := 0 to nrofbytes - 1 do
  begin
    writeByte(address, getByte(Value, i));
    address := address + 1;
  end;
end;

// =========================================
//               WRITING OUTPUT
// =========================================

procedure printTerrainType(TerrainType: Byte; HexMode: Boolean);
var
  i: Integer;
  entryFound: Boolean;
begin
  if HexMode then
    Write(toString(TerrainType))
  else
  begin
    entryFound:=False;
    for i := low(TerrainTypes) to high(TerrainTypes) do
    begin
      if TerrainTypes[i].ByteId = TerrainType then
      begin
        Write(TerrainTypes[i].Letter);
        entryFound:=true;
        Break;
      end;
    end;
    if not entryFound then
      Write('?');
  end;
end;

procedure printFileTerrainTypes(startIndex, length:Integer; HexMode: Boolean);
var
  i: Integer;
begin
  if FileSize(fileobj) < (startIndex + length) then
    length := startIndex - FileSize(fileobj);

  for i := 0 to length-1 do
  begin
    printTerrainType(readByte(startIndex + i), HexMode);
  end;
end;

procedure printTerrainTypes(cutoff:Integer; HexMode: Boolean);
var
  i: Integer;
begin
  cutoff:=Min(TerrainInputCellsLength, cutoff) - 1;
  for i := 0 to cutoff do
  begin
    printTerrainType(TerrainInputCells[i], HexMode);
  end;
end;

// Sets the variables "Cells", "CellsX", "CellsY" and "TypesOffset".
function openAndCheckFile(filename: String; IsCellsInputFile: Boolean): Boolean;
var
  size: Int64;
  xyproduct: Integer;
  errorMsg: String;
begin
  if IsCellsInputFile then
    errorMsg:='Input file error: '
  else
    errorMsg:='Error: ';
  Result:=openFile(filename);
  if not Result then
     Exit; // open file fail itself already gives error message
  size:=getFileSize();
  Result:=size > $28;
  if not Result then
  begin
    WriteLn(errorMsg + 'file is too short to contain a TMP header');
     Exit;
  end;
  Result:=(readBytes($0,2) = 24) AND (readBytes($2,2) = 24);
  if not Result then
  begin
    WriteLn(errorMsg + 'cell size is not 24x24. This is not a valid RA1 TMP file');
    Exit;
  end;
  Cells:=readBytes($4,4);
  CellsX:=readBytes($8,2);
  CellsY:=readBytes($A,2);
  xyproduct:= Cellsx * Cellsy;
  Result:= (xyproduct = Cells) OR (xyproduct = 1);
  if not Result then
  begin
    WriteLn(errorMsg + 'amount of cells and set dimensions don''t match!');
    Exit;
  end;
  TypesOffset:=readBytes($20,4);
  Result:=CellsX*CellsY <= High(TerrainInputCells);
  if not Result then
  begin
    WriteLn(errorMsg + 'this program can''t handle files with more than ' + ToString(High(TerrainInputCells)) + ' cells!');
    Exit;
  end;
  Result:=size > TypesOffset;
  if not Result then
    WriteLn(errorMsg + 'is too short to contain the type information!');
end;

procedure WriteTileInfo();
var
   IsSingleCell: Boolean;
   i, origlen: Integer;
   origc: Cardinal;
   origX, origY: Word;
begin
  if RunMode = 1 then
  begin
    origc := Cells;
    origX := CellsX;
    origY := cellsY;
  end;
  // test if file is a valid RA1 TMP file
  if not openAndCheckFile(TerrainOutputFile,false) then
    Exit;

  IsSingleCell:= (CellsX = 1) and (CellsY = 1) and (Cells > 1);
  if IsSingleCell then origlen:=1 else origlen:=Cells;
  // check if forcelength is needed to proceed, and pad array with defaults
  if (not FixToSingleCell) and (TerrainInputCellsLength < Cells) and ForceLength then
  begin
    for i:=TerrainInputCellsLength to Cells do
      TerrainInputCells[i]:=DefaultFill;
    TerrainInputCellsLength:=Cells;
  end;
  if RunMode <> 4 then
    write('Output f')
  else
    write('F');
  WriteLn('ile info:');
  if RunMode <> 4 then
    Write('-------');
  WriteLn('---------');
  WriteLn('Number of cells:    ' + toString(Cells));
  WriteLn('Tileset dimensions: ' + toString(CellsX) + 'x' + toString(CellsY));
  Write  ('Is 1x1-multiple:    ');
  if IsSingleCell then WriteLn('Yes') else WriteLn('No');
  Write  ('Terrain type mask:  ');
  printFileTerrainTypes(TypesOffset,origlen,false);
  WriteLn();
  if RunMode = 4 then
     Exit;
  WriteLn();
  if (not FixToSingleCell)
     and (Cells <> TerrainInputCellsLength)
     and (not ForceLength) then
  begin
    writeln('ERROR: Amount of cells in input and output does not match. Use the ''!'' modifier');
    writeln('to ignore this, and use ''-d'' to speficy a default to fill with.');
    closeFile();
    Exit;
  end;
  // only file import can restore dimensions
  if (IsSingleCell) and (RunMode <> 1) then
    TerrainInputCellsLength:=1;

  WriteLn('Edit info:');
  WriteLn('---------');
  if FixToSingleCell then
    WriteLn('-Converting to 1x1 with multiple entries.')
  else if (origc = Cells) and IsSingleCell then
  WriteLn('-Importing dimensions:  ' + toString(origX) + 'x' + toString(origY));
  Write  ('-New terrain type mask: ');
  printTerrainTypes(TerrainInputCellsLength,False);
  WriteLn();

  // write stuff to file
  //{
  WriteLn();
  Write('Changing file...');
  if (origc = Cells) and IsSingleCell and (not FixToSingleCell) then
  begin
    writeBytes($08,origX,2);
    writeBytes($0A,origY,2);
  end;

  if (FixToSingleCell) and (not IsSingleCell) then
  begin
    writeBytes($08,1,2);
    writeBytes($0A,1,2);
  end;
  seek(fileobj,TypesOffset);
  BlockWrite(fileobj,TerrainInputCells,TerrainInputCellsLength);
  if FilePos(fileobj) < FileSize(fileobj) then
     Truncate(fileobj);
  writeBytes($0C,FileSize(fileobj),4);
  //}
  closeFile();
  Writeln(' done.');
end;

// =========================================
//              PROCESSING INPUT
// =========================================

function getTerrainType(c: Char; HexRead: Boolean; var code: Integer): Byte;
var i: Integer;
  entryFound: boolean;
begin
  c := UpperCase(c)[1];
  if HexRead then
    Result:= hexchar2byte(c, code)
  else
  begin
    entryFound:=False;
    for i := low(TerrainTypes) to high(TerrainTypes) do
      if TerrainTypes[i].Letter = c then
      begin
        Result:=TerrainTypes[i].ByteId;
        entryFound:=true;
        Exit;
      end;
    if not entryFound then
      code:=1;
  end;
end;

procedure ProcessInputFile();
var
  inputFileExists: Boolean;
begin
  if TerrainInputFile = '' then
    WriteLn('Error: no terrain input filename given');
  inputFileExists:=FileExists(TerrainInputFile);
  if not inputFileExists then
  begin
    WriteLn('Error: terrain input file does not exist');
    RunMode:= -1;
    Exit;
  end;
  if not openAndCheckFile(TerrainInputFile,true) then
  begin
    RunMode:=5;
    Exit;
  end;
  seek(fileobj,TypesOffset);
  blockread (fileobj,TerrainInputCells,Cells,TerrainInputCellsLength);
  FixToSingleCell:= (CellsX = 1) and (CellsY = 1) and (Cells > 1);
  closeFile();
end;

procedure ProcessInputString();
var
  i, code: integer;
begin
  code:=0;
  if TerrainInputString = '' then
  begin
    WriteLn('Error: no terrain types given');
    RunMode:= -1;
    Exit;
  end;
  TerrainInputCellsLength:=Length(TerrainInputString);
  for i:=1 to TerrainInputCellsLength do
  begin
    TerrainInputCells[i-1] := getTerrainType(TerrainInputString[i], TerrainHexReadMode, code);
    if code = 1 then
    begin
      WriteLn('Error: Illegal character(s) in terrain types string.');
      RunMode:= -1;
      TerrainInputCellsLength:=0;
      exit;
    end;
  end;
end;

procedure ProcessParams();
var
  code: Integer;
begin
  if RunMode = 1 then
    ProcessInputFile()
  else if RunMode = 2 then
   ProcessInputString()
  else if RunMode = 3 then
  begin
    FixToSingleCell:=True;
    if TerrainInputString <> '' then
      ProcessInputString()
    else
    begin
      TerrainInputCellsLength:=1;
      TerrainInputCells[0]:=0;
    end;
  end;

  if DefaultFillInputChar <> Char(00) then
  begin
    code:=0;
    DefaultFill:=getTerrainType(DefaultFillInputChar,DefaultHexReadMode, code);
    if code <> 0 then
    begin
      WriteLn('Error: Illegal character given for default terrain type.');
      RunMode:= -1;
      exit;
    end;
  end;
end;

// =========================================
//                HELP SCREENS
// =========================================

procedure showHelp();
begin
  writeln('Basic syntax:');
  writeln('  RTMType file -[t|f|1][x][!] [arguments] -d[x] [defaultval]');
  writeln();
  writeln('Parameters:');
  writeln(' Types reading:');
  writeln('  t: Give terrain type codes as arguments');
  writeln('  f: Read terrain types from file given as argument');
  writeln('  1: Force tile to 1x1 with multiple entries. Terrain type is X/0 if not given.');
  writeln(' Modifiers:');
  writeln('  x : Read byte values (0-F) as codes.');
  writeln('  ! : Don''t abort if amount of cells in file and in input don''t match.');
  writeln(' Default value:');
  writeln('  d: Specify default value to fill with when using ''!'' mode. (Can use as -dx)');
  writeln('     If not given, this defaults to Clear (3). Is never needed when using ''-1''.');
  writeln();
  writeln('Start with -l to list all terrain types and their codes.');
  writeln('Start with -x to show explained examples.');
end;

procedure showCodes();
var
  i:integer;
begin
  WriteLn('Terrain types listing:');
  WriteLn('-------------------------------------------------------------------------------');
  WriteLn(' Code  Byte  Description');
  WriteLn('-------------------------------------------------------------------------------');
  for i := low(TerrainTypes) to high(TerrainTypes) do
  begin
    Write('  ' + TerrainTypes[i].Letter + '    ');
    Write(' ' + byte2hexchar(TerrainTypes[i].ByteId) + '    ');
    WriteLn(TerrainTypes[i].Desc);
  end;
  WriteLn('-------------------------------------------------------------------------------');
end;

procedure showExamples();
begin
  writeln('Examples:');
  writeln();
  writeln(' RTMType hill01.tem -t HIIIHIHHIIHC');
  writeln('   Set the terrain types in ''hill01.tem'' to the specified type codes.');
  writeln();
  writeln(' RTMType hill01.tem -tx e888e8ee88e3');
  writeln('   Set the terrain types in ''hill01.tem'' to the specified type bytes.');
  writeln('   Exactly the same effect as above, but can hack in normally-unused types.');
  writeln();
  writeln(' RTMType hill01.tem -f! xtra0014.int -d I');
  writeln('   Set the terrain types in ''hill01.tem'' to the ones set in ''xtra0014.int''.');
  writeln('   Continue even if the amount of cells doesn''t match. If amount of cells in');
  writeln('   ''xtra0014.int'' is less than in ''hill01.tem'', pad it with ''Impassable'' type.');
  writeln('   Note: ''!'' is not needed if the types are read from a 1x1-multiple file.');
  writeln();
  writeln(' RTMType arro0001.int -1');
  writeln('   Set the given tile to act as 1x1 tile with multiple entries. The terrain');
  writeln('   type defaults to 0 here since no type codes/bytes are given behind the ''-1''.');
end;

  // =========================================
  //               INITIALIZATION
  // =========================================

// Checks all command line parameters for relevant information and saves the found information.
procedure ReadParams();
var
  Param: String;
  ParamId, ParamMod1, ParamMod2: Char;
  ParamLen, i:integer;
  AcceptParam: boolean;
begin
  TerrainOutputFile:=ParamStr(1);
  if not FileExists(TerrainOutputFile) then
  begin
    writeln('Error: Input file ' + TerrainOutputFile + ' not found!');
    RunMode := -1;
    Exit;
  end
  else
    RunMode := 4;
  paramId:=' ';
  paramMod1:=' ';
  paramMod2:=' ';
  AcceptParam:=False;
  for i:=2 to ParamCount do
  begin
    param:=ParamStr(i);
    paramLen:=Length(param);
    if ((not acceptparam) or (paramId = ' ')) then
    begin // read actual parameter and modifiers
      if (param[1] = '-') and ( paramLen > 1) then
      begin
        // read parameter info
        AcceptParam:=true;
        paramId:= LowerCase(param[2]);
        if (paramLen > 2) then
          paramMod1:=LowerCase(param[3]);
        if (paramLen > 3) then
          paramMod2:=LowerCase(param[4]);

        // process parameter info
        if      (paramId = 'f') then // use input file
        begin
          RunMode:=1;
          ForceLength:=ParamMod1 = '!';
        end
        else if (paramId = 't') then // read types from string
        begin
          RunMode:=2;
          if ParamMod1 = 'x' then
          begin
            TerrainHexReadMode:=True;
            ForceLength:= ParamMod2 = '!';
          end
          else
            ForceLength:=ParamMod1 = '!';
        end
        else if (paramId = '1') then // Force to 1x1 repeating
        begin
          RunMode:=3;
          TerrainHexReadMode:=ParamMod1 = 'x';
        end
        else if (paramId = 'd') then // read default
          DefaultHexReadMode:=ParamMod1 = 'x'
        else                         // abort if illegal parameter is found
        begin
          RunMode := ($100 + byte(paramId)) * -1;
          Exit;
        end;
      end;
    end
    else
    begin // read argument of read parameter
      if paramId = 'f' then
        TerrainInputFile:=param
      else if paramId = 't' then
        TerrainInputString:=Param
      else if paramId = '1' then
        TerrainInputString:=Param[1]
      else if paramId = 'd' then
      begin
        DefaultFillInputChar:=Param[1];
      end;
      AcceptParam:=false;
      paramId:=' ';
      paramMod1:=' ';
      paramMod2:=' ';
    end
  end;
  {
  if (RunMode = 0) then
  begin
    writeln('Error: No terrain type input parameters given.');
    RunMode := -1;
    Exit;
  end;
  }
end; {checkparams}


procedure ShowParamError();
var
  decodedChar: integer;
begin
  decodedChar:=(RunMode * -1) and $FF;
  writeln('Error: Illegal parameter ''' + Char(decodedChar) + '''');
end;

// returns False if any help parameters were found and handled.
function HandleHelpParams(): Boolean;
var firstParam: String;
begin
  result:= False;
  firstParam:= UpperCase(ParamStr(1));
  if (firstParam = '-H') or (firstParam = '/H') or (firstParam = '-?') or (firstParam = '/?') then
    showHelp()
  else if firstParam = '-L' then
    showCodes()
  else if firstParam = '-X' then
    showExamples()
  else
    result:= True;
end;

procedure ClearDefaults();
begin
  // internal info
  RunMode:=0;
  TerrainOutputFile:='';
  TerrainInputCellsLength:=0;
  ForceLength:=False;
  DefaultFill:=$03;
  FixToSingleCell:=false;
  // parameter info. Stops being relevant after the internal options are all set.
  TerrainInputFile:='';
  TerrainInputString:='';
  TerrainHexReadMode:=False;
  DefaultFillInputChar:=Char($00);
  DefaultHexReadMode:=False;
end;

{$R *.res}

begin
  //ClrScr();
  WriteLn('Red Alert Template File Terrain Type Editing Tool');
  WriteLn(' Created by Nyerguds');
  WriteLn('');
  if ParamCount > 0 then
  begin
    if HandleHelpParams() then
    begin
      ClearDefaults();
      ReadParams();
      if RunMode < -$100 then
        ShowParamError()
      else if RunMode = 0 then
        showHelp()
      else
      begin
        ProcessParams();
        // runmode 0 = show help
        // runmode 5 = input error; abort
        if (RunMode > 0) and (RunMode <> 5) then
           WriteTileInfo();
      end;
      if RunMode <1 then
      begin
        WriteLn();
        WriteLn('Start with ''-h'' for help.');
      end;
    end;
  end
  else
  begin
    showHelp();
    cntin(True);
  end;
end.
